"use strict";
/*
 * Copyright (C) 2021 William Stein <wstein@sagemath.com>
 * Copyright (C) 2015 Kovid Goyal <kovid at kovidgoyal.net>
 *
 * Distributed under terms of the BSD license
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var path_1 = require("path");
var fs_1 = require("fs");
var promises_1 = require("fs/promises");
var vm_1 = require("vm");
var utils_1 = require("./utils");
var compiler_1 = __importDefault(require("./compiler"));
var JPython = (0, compiler_1.default)();
// Async because also capable of reading to EOF from stdin.
function readWholeFile(filename) {
    return __awaiter(this, void 0, void 0, function () {
        var chunks;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    if (!filename) return [3 /*break*/, 2];
                    return [4 /*yield*/, (0, promises_1.readFile)(filename)];
                case 1: return [2 /*return*/, (_a.sent()).toString()];
                case 2:
                    chunks = [];
                    process.stdin.setEncoding("utf-8");
                    process.stdin.on("data", function (data) {
                        chunks.push(data.toString());
                    });
                    process.openStdin();
                    return [4 /*yield*/, (0, utils_1.once)(process.stdin, "end")];
                case 3:
                    _a.sent();
                    return [2 /*return*/, chunks.join("")];
            }
        });
    });
}
function process_cache_dir(dir) {
    dir = (0, path_1.resolve)((0, path_1.normalize)(dir));
    (0, fs_1.mkdirSync)(dir, { recursive: true });
    return dir;
}
function Compile(_a) {
    var argv = _a.argv, src_path = _a.src_path, lib_path = _a.lib_path;
    return __awaiter(this, void 0, void 0, function () {
        function parseFile(code, filename) {
            return JPython.parse(code, {
                filename: filename,
                basedir: filename !== "<stdin>" ? (0, path_1.dirname)(filename) : undefined,
                libdir: (0, path_1.join)(src_path, "lib"),
                import_dirs: (0, utils_1.getImportDirs)(argv.import_path),
                discard_asserts: argv.discard_asserts,
                module_cache_dir: module_cache_dir,
                sage: argv.sage,
            });
        }
        function writeOutput(output) {
            if (argv.output) {
                if (argv.output == "/dev/stdout") {
                    // Node's filesystem module doesn't write directly to /dev/stdout
                    console.log(output);
                }
                else if (argv.output == "/dev/stderr") {
                    console.error(output);
                }
                else {
                    (0, fs_1.writeFileSync)(argv.output, output, "utf8");
                }
            }
            else if (!argv.execute) {
                console.log(output);
            }
            if (argv.execute) {
                // @ts-ignore
                global.require = require;
                (0, vm_1.runInThisContext)(output);
            }
        }
        function timeIt(name, f) {
            var t1 = new Date().getTime();
            f();
            if (argv.stats) {
                var spent = new Date().getTime() - t1;
                if (stats[name]) {
                    stats[name] += spent;
                }
                else {
                    stats[name] = spent;
                }
            }
        }
        function compileSingleFile(code) {
            return __awaiter(this, void 0, void 0, function () {
                var topLevel, output;
                return __generator(this, function (_a) {
                    timeIt("parse", function () {
                        var filename = files[0] || argv.filename_for_stdin || "<stdin>";
                        try {
                            topLevel = parseFile(code, filename);
                        }
                        catch (err) {
                            if (!(err instanceof JPython.SyntaxError)) {
                                throw err;
                            }
                            console.error(err.toString());
                            process.exit(1);
                        }
                    });
                    try {
                        output = new JPython.OutputStream(outputOptions);
                    }
                    catch (err) {
                        if (err instanceof JPython.DefaultsError) {
                            console.error(err.message);
                            process.exit(1);
                        }
                        throw err;
                    }
                    timeIt("generate", function () {
                        topLevel.print(output);
                    });
                    output = output.get();
                    writeOutput(output);
                    return [2 /*return*/];
                });
            });
        }
        var module_cache_dir, outputOptions, files, stats, count, files_1, files_1_1, filename, _b, e_1_1, _c, name_1;
        var e_1, _d;
        return __generator(this, function (_e) {
            switch (_e.label) {
                case 0:
                    module_cache_dir = argv.cache_dir
                        ? process_cache_dir(argv.cache_dir)
                        : "";
                    outputOptions = {
                        beautify: true,
                        private_scope: !argv.bare,
                        omit_baselib: argv.omit_baselib,
                        keep_docstrings: argv.keep_docstrings,
                        discard_asserts: argv.discard_asserts,
                        module_cache_dir: module_cache_dir,
                    };
                    files = argv.files.slice();
                    stats = {};
                    count = files.length || 1;
                    if (argv.comments) {
                        if (/^\//.test(argv.comments)) {
                            outputOptions.comments = new Function("return(" + argv.comments + ")")();
                        }
                        else if (argv.comments == "all") {
                            outputOptions.comments = true;
                        }
                        else {
                            outputOptions.comments = function (_, comment) {
                                var value = comment.value;
                                var type = comment.type;
                                if (type == "comment2") {
                                    // multiline comment
                                    return /@preserve|@license|@cc_on/i.test(value);
                                }
                            };
                        }
                    }
                    if (!argv.omit_baselib) {
                        outputOptions.baselib_plain = (0, fs_1.readFileSync)((0, path_1.join)(lib_path, "baselib-plain-pretty.js"), "utf-8");
                    }
                    if (files.filter(function (el) { return el == "-"; }).length > 1) {
                        console.error("ERROR: Can only read a single file from STDIN (two or more dashes specified)");
                        process.exit(1);
                    }
                    if (!(files.length > 0)) return [3 /*break*/, 10];
                    _e.label = 1;
                case 1:
                    _e.trys.push([1, 7, 8, 9]);
                    files_1 = __values(files), files_1_1 = files_1.next();
                    _e.label = 2;
                case 2:
                    if (!!files_1_1.done) return [3 /*break*/, 6];
                    filename = files_1_1.value;
                    _b = compileSingleFile;
                    return [4 /*yield*/, readWholeFile(filename)];
                case 3: return [4 /*yield*/, _b.apply(void 0, [_e.sent()])];
                case 4:
                    _e.sent();
                    _e.label = 5;
                case 5:
                    files_1_1 = files_1.next();
                    return [3 /*break*/, 2];
                case 6: return [3 /*break*/, 9];
                case 7:
                    e_1_1 = _e.sent();
                    e_1 = { error: e_1_1 };
                    return [3 /*break*/, 9];
                case 8:
                    try {
                        if (files_1_1 && !files_1_1.done && (_d = files_1.return)) _d.call(files_1);
                    }
                    finally { if (e_1) throw e_1.error; }
                    return [7 /*endfinally*/];
                case 9: return [3 /*break*/, 13];
                case 10:
                    _c = compileSingleFile;
                    return [4 /*yield*/, readWholeFile()];
                case 11: return [4 /*yield*/, _c.apply(void 0, [_e.sent()])];
                case 12:
                    _e.sent();
                    _e.label = 13;
                case 13:
                    if (argv.stats) {
                        console.error("Timing information (compressed ".concat(count, " files):"));
                        for (name_1 in stats)
                            console.error("- ".concat(name_1, ": ").concat((stats[name_1] / 1000).toFixed(3), "s"));
                    }
                    return [2 /*return*/];
            }
        });
    });
}
exports.default = Compile;
//# sourceMappingURL=compile.js.map