"use strict";
/*
 * Copyright (C) 2021 William Stein <wstein@sagemath.com>
 * Copyright (C) 2015 Kovid Goyal <kovid at kovidgoyal.net>
 *
 * Distributed under terms of the BSD license
 */
Object.defineProperty(exports, "__esModule", { value: true });
// Thin wrapper around (release|dev)/compiler.js to setup some global facilities and
// export the compiler's symbols safely.
var path_1 = require("path");
var fs_1 = require("fs");
var vm_1 = require("vm");
var utils_1 = require("./utils");
function createCompiler(options) {
    var _a;
    if (options === void 0) { options = {}; }
    var compiler_exports = {};
    var compiler_context = (0, vm_1.createContext)({
        console: (_a = options.console) !== null && _a !== void 0 ? _a : console,
        readfile: fs_1.readFileSync,
        writefile: fs_1.writeFileSync,
        sha1sum: utils_1.sha1sum,
        require: require,
        exports: compiler_exports,
    });
    var base = (0, path_1.join)(__dirname, "..", "..");
    var compiler_dir = (0, path_1.join)(base, "dev");
    if (!(0, utils_1.pathExists)((0, path_1.join)(compiler_dir, "compiler.js"))) {
        compiler_dir = (0, path_1.join)(base, "release");
    }
    var compiler_file = (0, path_1.join)(compiler_dir, "compiler.js");
    var compilerjs = (0, fs_1.readFileSync)(compiler_file, "utf-8");
    (0, vm_1.runInContext)(compilerjs, compiler_context, (0, path_1.relative)(base, compiler_file));
    return compiler_exports;
}
exports.default = createCompiler;
//# sourceMappingURL=compiler.js.map