"use strict";
/*
 * Copyright (C) 2021 William Stein <wstein@sagemath.com>
 * Copyright (C) 2015 Kovid Goyal <kovid at kovidgoyal.net>
 *
 * Distributed under terms of the BSD license
 */
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
Object.defineProperty(exports, "__esModule", { value: true });
var vm_1 = require("vm");
function Completer(compiler) {
    var allKeywords = compiler.ALL_KEYWORDS.split(" ");
    function globalNames() {
        try {
            var names = (0, vm_1.runInThisContext)("Object.getOwnPropertyNames((() => this)())");
            return __spreadArray([], __read(new Set(names.concat(allKeywords))), false).sort();
        }
        catch (e) {
            console.log(e.stack || e.toString());
        }
        return [];
    }
    function objectNames(obj, prefix) {
        var _a;
        if (obj == null)
            return [];
        var names = [];
        function add(o) {
            var items = Object.getOwnPropertyNames(o).filter(function (name) {
                return name.startsWith(prefix);
            });
            names.push.apply(names, __spreadArray([], __read(items), false));
        }
        var p;
        if (typeof obj === "object" || typeof obj === "function") {
            add(obj);
            p = Object.getPrototypeOf(obj);
        }
        else {
            p = (_a = obj.constructor) === null || _a === void 0 ? void 0 : _a.prototype;
        }
        // Walk the prototype chain
        try {
            // Circular refs possible? Let's guard against that.
            for (var sentinel = 0; sentinel < 5 && p != null; sentinel++) {
                add(p);
                p = Object.getPrototypeOf(p);
            }
        }
        catch (_err) { }
        // unique and sorted:
        return __spreadArray([], __read(new Set(names)), false).sort();
    }
    function prefixMatches(prefix, items) {
        return items.filter(function (item) { return item.startsWith(prefix); }).sort();
    }
    function findCompletions(line) {
        var e_1, _a;
        var t;
        try {
            t = compiler.tokenizer(line, "<repl>");
        }
        catch (_err) {
            return [];
        }
        var tokens = [];
        var token;
        while (true) {
            try {
                token = t();
            }
            catch (_err) {
                return [];
            }
            if (token.type === "eof")
                break;
            if (token.type === "punc" && "(){},;:".includes(token.value)) {
                // empties the tokens since we care about what's next only
                tokens.splice(0, tokens.length);
            }
            tokens.push(token);
        }
        if (tokens.length == 0) {
            // New line or trailing space
            return [globalNames(), ""];
        }
        var lastTok = tokens[tokens.length - 1];
        if (lastTok.value === "." ||
            (lastTok.type === "name" && compiler.IDENTIFIER_PAT.test(lastTok.value))) {
            lastTok = lastTok.value;
            if (lastTok === ".") {
                tokens.push({ value: "" });
                lastTok = "";
            }
            if (tokens.length > 1 && tokens[tokens.length - 2].value === ".") {
                // A compound expression
                var prefix = "";
                var result = void 0;
                try {
                    for (var _b = __values(tokens.slice(0, tokens.length - 2)), _c = _b.next(); !_c.done; _c = _b.next()) {
                        var tok = _c.value;
                        prefix += tok.value;
                    }
                }
                catch (e_1_1) { e_1 = { error: e_1_1 }; }
                finally {
                    try {
                        if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
                    }
                    finally { if (e_1) throw e_1.error; }
                }
                if (prefix) {
                    try {
                        result = (0, vm_1.runInThisContext)(prefix);
                    }
                    catch (e) {
                        return [];
                    }
                    return [objectNames(result, lastTok), lastTok];
                }
            }
            else {
                return [prefixMatches(lastTok, globalNames()), lastTok];
            }
        }
        return [];
    }
    return findCompletions;
}
exports.default = Completer;
//# sourceMappingURL=completer.js.map