"use strict";
/*
 * Copyright (C) 2021 William Stein <wstein@sagemath.com>
 * Copyright (C) 2015 Kovid Goyal <kovid at kovidgoyal.net>
 *
 * Distributed under terms of the BSD license
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _a;
Object.defineProperty(exports, "__esModule", { value: true });
var fs_1 = require("fs");
var path_1 = require("path");
var vm_1 = require("vm");
var utils_1 = require("./utils");
var completer_1 = __importDefault(require("./completer"));
var readline_1 = require("readline");
var compiler_1 = __importDefault(require("./compiler"));
var DEFAULT_HISTORY_SIZE = 1000;
var HOME = (_a = process.env[process.platform == "win32" ? "USERPROFILE" : "HOME"]) !== null && _a !== void 0 ? _a : "/tmp";
function expandUser(x) {
    return x.replace("~", HOME);
}
var CACHEDIR = process.env.XDG_CACHE_HOME
    ? expandUser(process.env.XDG_CACHE_HOME)
    : (0, path_1.join)(HOME, ".cache");
function replDefaults(options) {
    var _a, _b;
    if (!options.input) {
        options.input = process.stdin;
    }
    if (!options.output) {
        options.output = process.stdout;
    }
    if (options.show_js == null) {
        options.show_js = true;
    }
    if (!options.ps1) {
        if (options.jsage) {
            options.ps1 = process.stdin.isTTY ? "jsage: " : "";
        }
        else {
            options.ps1 = process.stdin.isTTY ? ">>> " : "";
        }
    }
    if (!options.ps2) {
        options.ps2 = process.stdin.isTTY ? "... " : "";
    }
    if (!options.console) {
        options.console = console;
    }
    if (options.terminal == null) {
        options.terminal = !!((_a = options.output) === null || _a === void 0 ? void 0 : _a.isTTY);
    }
    if (options.histfile == null) {
        var CACHE = (0, path_1.join)(CACHEDIR, "jpython");
        if (!(0, utils_1.pathExists)(CACHE)) {
            (0, fs_1.mkdirSync)(CACHE, { recursive: true });
        }
        options.histfile = (0, path_1.join)(CACHE, "history");
    }
    options.historySize = (_b = options.historySize) !== null && _b !== void 0 ? _b : DEFAULT_HISTORY_SIZE;
    return options;
}
function readHistory(options) {
    if (options.histfile) {
        if (!(0, utils_1.pathExists)(options.histfile)) {
            return [];
        }
        try {
            return (0, fs_1.readFileSync)(options.histfile, "utf-8").split("\n");
        }
        catch (err) {
            options.console.warn("Error reading history file - ".concat(err));
            return [];
        }
    }
    return [];
}
function writeHistory(options, history) {
    if (options.histfile) {
        try {
            return (0, fs_1.writeFileSync)(options.histfile, history.join("\n"), "utf-8");
        }
        catch (err) {
            options.console.warn("Error writing history file - ".concat(err));
        }
    }
}
function createReadlineInterface(options, JPython) {
    var _a;
    // See https://nodejs.org/api/readline.html#readline_readline_createinterface_options
    var completer = (0, completer_1.default)(JPython);
    var history = options.terminal ? readHistory(options) : [];
    var readline = ((_a = options.mockReadline) !== null && _a !== void 0 ? _a : readline_1.createInterface)({
        input: options.input,
        output: options.output,
        completer: completer,
        terminal: options.terminal,
        history: history,
        historySize: options.historySize,
        tabSize: 4,
    });
    // @ts-ignore -- needed for older node.js
    readline.history = history;
    return readline;
}
function Repl(options0) {
    return __awaiter(this, void 0, void 0, function () {
        function duringInit(line) {
            initLines.push(line);
        }
        function printAST(ast, keepBaselib) {
            var output = new JPython.OutputStream({
                omit_baselib: !keepBaselib,
                write_name: false,
                private_scope: false,
                beautify: true,
                keep_docstrings: true,
                baselib_plain: keepBaselib
                    ? (0, fs_1.readFileSync)((0, path_1.join)(utils_1.libraryPath, "baselib-plain-pretty.js"), "utf-8")
                    : undefined,
            });
            ast.print(output);
            return output.get();
        }
        function initContext() {
            return __awaiter(this, void 0, void 0, function () {
                var BLOCK, jsage;
                return __generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            // @ts-ignore
                            global.require = require;
                            // and get all the code and name.
                            (0, vm_1.runInThisContext)(printAST(JPython.parse("(def ():\n yield 1\n)"), true));
                            (0, vm_1.runInThisContext)('var __name__ = "__repl__"; show_js=false;');
                            if (!options.jsage) return [3 /*break*/, 3];
                            BLOCK = true;
                            if (!BLOCK) return [3 /*break*/, 2];
                            jsage = require("@jsage/lib");
                            return [4 /*yield*/, jsage.init()];
                        case 1:
                            _a.sent();
                            _a.label = 2;
                        case 2:
                            (0, vm_1.runInThisContext)("jsage = require('@jsage/lib');");
                            (0, vm_1.runInThisContext)("for(const x in jsage) { global[x] = jsage[x]; }");
                            _a.label = 3;
                        case 3: return [2 /*return*/];
                    }
                });
            });
        }
        function resetBuffer() {
            buffer.splice(0, buffer.length);
        }
        function prompt() {
            var leadingWhitespace = "";
            if (more && buffer.length) {
                var prev_line = buffer[buffer.length - 1];
                if (prev_line.trimRight().slice(-1) == ":") {
                    leadingWhitespace = "    ";
                }
                // Add to leadingWhitespace all the blank space at the beginning of prev_line, if any.
                var match = prev_line.match(/^\s+/);
                if (match) {
                    leadingWhitespace += match[0];
                }
            }
            readline.setPrompt(more ? ps2 : ps1);
            readline.prompt();
            if (leadingWhitespace) {
                readline.write(leadingWhitespace);
            }
        }
        function runJS(js, noPrint) {
            if ((0, vm_1.runInThisContext)("show_js")) {
                options.console.log(colorize("---------- Compiled JavaScript ---------", "green", true));
                options.console.log(js);
                options.console.log(colorize("---------- Running JavaScript ---------", "green", true));
            }
            var result;
            try {
                global.console = options.console;
                result = (0, vm_1.runInThisContext)(js);
            }
            catch (err) {
                if (err === null || err === void 0 ? void 0 : err.stack) {
                    options.console.error(err === null || err === void 0 ? void 0 : err.stack);
                }
                else {
                    options.console.error(err);
                }
            }
            if (!noPrint && result != null && global.ρσ_print != null) {
                // We just print out the last result using normal Python printing.
                try {
                    global.ρσ_print(result);
                }
                catch (err) {
                    if (err === null || err === void 0 ? void 0 : err.stack) {
                        options.console.error(err === null || err === void 0 ? void 0 : err.stack);
                    }
                    else {
                        options.console.error(err);
                    }
                }
            }
        }
        // returns true if incomplete
        function compileAndRun(source) {
            var time = undefined;
            if (source.startsWith("%time ") || source.startsWith("time ")) {
                time = 0;
                source = source.slice(5).trimLeft();
            }
            var classes = toplevel === null || toplevel === void 0 ? void 0 : toplevel.classes;
            var scoped_flags = toplevel === null || toplevel === void 0 ? void 0 : toplevel.scoped_flags;
            try {
                toplevel = JPython.parse(source, {
                    filename: "<repl>",
                    basedir: process.cwd(),
                    libdir: utils_1.importPath,
                    import_dirs: importDirs,
                    classes: classes,
                    scoped_flags: scoped_flags,
                    jsage: options.jsage,
                    tokens: options.tokens,
                });
            }
            catch (err) {
                if (err.is_eof && err.line == buffer.length && err.col > 0) {
                    return true;
                }
                if (err.message && err.line !== undefined) {
                    options.console.log(err.line + ":" + err.col + ":" + err.message);
                }
                else {
                    options.console.log(err.stack || err);
                }
                return false;
            }
            var output = printAST(toplevel);
            if (classes) {
                var exports_1 = {};
                for (var name_1 in toplevel.exports) {
                    exports_1[name_1] = true;
                }
                for (var name_2 in classes) {
                    if (!exports_1[name_2] && !toplevel.classes[name_2]) {
                        toplevel.classes[name_2] = classes[name_2];
                    }
                }
            }
            var noPrint = source.trimRight().endsWith(";");
            if (time != null) {
                time = new Date().valueOf();
            }
            runJS(output, noPrint);
            if (time) {
                console.log("Wall time: ".concat(new Date().valueOf() - time, "ms"));
            }
            return false;
        }
        // returns true if incomplete
        function push(line) {
            buffer.push(line);
            var trimmedLine = line.trimRight();
            if (trimmedLine &&
                LINE_CONTINUATION_CHARS.includes(trimmedLine.slice(-1))) {
                // ends in continuation character after trimming whitespace
                return true;
            }
            var source = buffer.join("\n");
            if (!source.trim()) {
                // all whitespace
                resetBuffer();
                return false;
            }
            var isIncomplete = compileAndRun(source);
            if (!isIncomplete) {
                resetBuffer();
            }
            return isIncomplete;
        }
        function readLine(line) {
            if (more) {
                // We are in a block
                var lineIsEmpty = !line.trimLeft();
                if (lineIsEmpty && buffer.length > 0) {
                    // We have an empty lines, so evaluate the block:
                    more = push(line.trimLeft());
                }
                else {
                    buffer.push(line);
                }
            }
            else {
                // Not in a block, evaluate line
                more = push(line);
            }
            prompt();
        }
        var options, JPython, readline, colorize, ps1, ps2, initLines, buffer, more, LINE_CONTINUATION_CHARS, toplevel, importDirs, initLines_1, initLines_1_1, line;
        var e_1, _a;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    options = replDefaults(options0);
                    JPython = (0, compiler_1.default)({
                        console: options.console,
                    });
                    readline = createReadlineInterface(options, JPython);
                    colorize = options.mockReadline
                        ? function (string, _color, _bold) { return string; }
                        : utils_1.colored;
                    ps1 = colorize(options.ps1, "blue");
                    ps2 = colorize(options.ps2, "green");
                    initLines = [];
                    readline.on("line", duringInit);
                    return [4 /*yield*/, initContext()];
                case 1:
                    _b.sent();
                    readline.off("line", duringInit);
                    buffer = [];
                    more = false;
                    LINE_CONTINUATION_CHARS = ":\\";
                    importDirs = (0, utils_1.getImportDirs)();
                    if (process.stdin.isTTY) {
                        options.console.log(colorize("Welcome to ".concat(options.jsage ? "JSage" : "JPython", ".  Using Node.js ").concat(process.version, ".  ").concat(options.jsage ? "\nType dir(jsage) for available functions." : ""), "green", true));
                    }
                    try {
                        // Run code we received during initialization.
                        for (initLines_1 = __values(initLines), initLines_1_1 = initLines_1.next(); !initLines_1_1.done; initLines_1_1 = initLines_1.next()) {
                            line = initLines_1_1.value;
                            readLine(line);
                        }
                    }
                    catch (e_1_1) { e_1 = { error: e_1_1 }; }
                    finally {
                        try {
                            if (initLines_1_1 && !initLines_1_1.done && (_a = initLines_1.return)) _a.call(initLines_1);
                        }
                        finally { if (e_1) throw e_1.error; }
                    }
                    readline.on("line", readLine);
                    readline.on("history", function (history) {
                        // Note -- this only exists in node >15.x.
                        if (options.terminal) {
                            writeHistory(options, history);
                        }
                    });
                    readline.on("close", function () {
                        var history = readline.history; //  deprecated in node 15...
                        if (history) {
                            writeHistory(options, history);
                        }
                        options.console.log();
                        process.exit(0);
                    });
                    readline.on("SIGINT", function () {
                        (0, readline_1.clearLine)(options.output, 0);
                        options.console.log("Keyboard Interrupt");
                        resetBuffer();
                        more = false;
                        prompt();
                    });
                    readline.on("SIGCONT", prompt);
                    prompt();
                    return [2 /*return*/];
            }
        });
    });
}
exports.default = Repl;
//# sourceMappingURL=repl.js.map