"use strict";
/*
 * Copyright (C) 2021 William Stein <wstein@sagemath.com>
 * Copyright (C) 2015 Kovid Goyal <kovid at kovidgoyal.net>
 *
 * Distributed under terms of the BSD license
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var path_1 = require("path");
var fs_1 = require("fs");
var compiler_1 = __importDefault(require("./compiler"));
var utils_1 = require("./utils");
var assert_1 = require("assert");
var os_1 = require("os");
var vm_1 = require("vm");
var JPython = (0, compiler_1.default)();
function default_1(argv, basePath, srcPath, libPath) {
    var e_1, _a;
    // run all tests and exit
    var failures = [];
    var compilerDir = (0, path_1.join)(basePath, "dev");
    if (!(0, utils_1.pathExists)((0, path_1.join)(compilerDir, "compiler.js"))) {
        compilerDir = (0, path_1.join)(basePath, "release");
    }
    var testPath = (0, path_1.join)(basePath, "test");
    var baselib = (0, fs_1.readFileSync)((0, path_1.join)(libPath, "baselib-plain-pretty.js"), "utf-8");
    var files = argv.files.length > 0
        ? argv.files.map(function (fname) {
            return fname.endsWith(".py") ? fname : fname + ".py";
        })
        : (0, fs_1.readdirSync)(testPath)
            .filter(function (name) { return /^[^_].*\.py$/.test(name); })
            .map(function (name) { return (0, path_1.join)(testPath, name); });
    var t_start = new Date().valueOf();
    var _loop_1 = function (filename) {
        var t0 = new Date().valueOf();
        var failed = false;
        var toplevel = undefined;
        try {
            var file = (0, fs_1.readFileSync)(filename, "utf-8");
            if (file.toString().includes("# DISABLED")) {
                console.log("Skipping ".concat(filename));
                return "continue";
            }
            toplevel = JPython.parse(file, {
                filename: filename,
                toplevel: toplevel,
                basedir: testPath,
                libdir: (0, path_1.join)(srcPath, "lib"),
            });
        }
        catch (err) {
            failures.push(filename);
            failed = true;
            console.log((0, utils_1.colored)(filename, "red") + ": " + err + "\n\n");
            return "break";
        }
        // generate output
        var output = new JPython.OutputStream({
            baselib_plain: baselib,
            beautify: true,
            keep_docstrings: true,
        });
        toplevel.print(output);
        // test that output performs correct JS operations
        var jsfile = (0, path_1.join)((0, os_1.tmpdir)(), (0, path_1.basename)(filename) + ".js");
        var code = output.toString();
        var assrt = __assign(__assign({}, require("assert")), { deepEqual: deepEqual });
        // We save and restore the console attributes since some tests,
        // e.g., repl, have a side effect of stealing them, which means
        // we suddenly can't report results.
        var saveConsole = __assign({}, console);
        var restoreConsole = function () {
            for (var name_1 in saveConsole) {
                console[name_1] = saveConsole[name_1];
            }
        };
        try {
            (0, vm_1.runInNewContext)(code, {
                assrt: assrt,
                __name__: jsfile,
                require: require,
                fs: require("fs"),
                JPython: JPython,
                console: console,
                compiler_dir: compilerDir,
                test_path: testPath,
                Buffer: Buffer,
            }, { filename: jsfile });
            restoreConsole();
        }
        catch (err) {
            restoreConsole();
            failures.push(filename);
            failed = true;
            (0, fs_1.writeFileSync)(jsfile, code);
            console.error("Failed running: " + (0, utils_1.colored)(jsfile, "red"));
            if (err.stack) {
                console.error((0, utils_1.colored)(filename, "red") + ":\n" + err.stack + "\n\n");
            }
            else {
                console.error((0, utils_1.colored)(filename, "red") + ": " + err + "\n\n");
            }
        }
        console.log("".concat((0, utils_1.colored)(filename, "green"), ": test ").concat(failed ? "FAILED" : "completed successfully", " (").concat(new Date().valueOf() - t0, "ms)"));
    };
    try {
        for (var files_1 = __values(files), files_1_1 = files_1.next(); !files_1_1.done; files_1_1 = files_1.next()) {
            var filename = files_1_1.value;
            var state_1 = _loop_1(filename);
            if (state_1 === "break")
                break;
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (files_1_1 && !files_1_1.done && (_a = files_1.return)) _a.call(files_1);
        }
        finally { if (e_1) throw e_1.error; }
    }
    if (failures.length > 0) {
        console.log((0, utils_1.colored)("There were " + failures.length + " test failure(s):", "red"));
        console.log.apply(console, failures);
    }
    else {
        console.log((0, utils_1.colored)("All tests passed! (".concat(new Date().valueOf() - t_start, "ms)"), "green"));
    }
    process.exit(failures.length ? 1 : 0);
}
exports.default = default_1;
// Modified version of deepEqual test assertion that is more suitable
// for testing python code.
function deepEqual(a, b, message) {
    if (Array.isArray(a) && Array.isArray(b)) {
        // Compare array objects that have extra properties as simple arrays
        if (a === b)
            return;
        if (a.length !== b.length)
            throw new assert_1.AssertionError({
                actual: a,
                expected: b,
                operator: "deepEqual",
                stackStartFn: deepEqual,
            });
        for (var i = 0; i < a.length; i++) {
            deepEqual(a[i], b[i], message);
        }
    }
    else if (typeof (a === null || a === void 0 ? void 0 : a.__eq__) === "function") {
        // Python operator overloading
        if (!a.__eq__(b))
            throw new assert_1.AssertionError({
                actual: a,
                expected: b,
                operator: "deepEqual",
                stackStartFn: deepEqual,
            });
    }
    else {
        // Fallback to standard version in nodejs library.
        return (0, assert_1.deepEqual)(a, b, message);
    }
}
//# sourceMappingURL=test.js.map