"use strict";
/*
 * Copyright (C) 2021 William Stein <wstein@sagemath.com>
 * Copyright (C) 2015 Kovid Goyal <kovid at kovidgoyal.net>
 *
 * Distributed under terms of the BSD license
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.once = exports.sha1sum = exports.getImportDirs = exports.merge = exports.wrap = exports.generators_available = exports.repeat = exports.colored = exports.pathExists = exports.comment_contents = exports.libraryPath = exports.importPath = exports.basePath = void 0;
var fs_1 = require("fs");
var path_1 = require("path");
var crypto_1 = require("crypto");
var events_1 = require("events");
var util_1 = require("util");
var path_2 = require("path");
exports.basePath = (0, path_2.normalize)((0, path_2.join)((0, path_2.dirname)(module.filename), "..", ".."));
exports.importPath = (0, path_2.join)(exports.basePath, "src", "lib");
exports.libraryPath = (0, path_2.join)(exports.basePath, "dev");
exports.comment_contents = /\/\*!?(?:\@preserve)?[ \t]*(?:\r\n|\n)([\s\S]*?)(?:\r\n|\n)[ \t]*\*\//;
var colors = ["red", "green", "yellow", "blue", "magenta", "cyan", "white"];
function ansi(code) {
    code = code || 0;
    return String.fromCharCode(27) + "[" + code + "m";
}
function pathExists(path) {
    try {
        (0, fs_1.statSync)(path);
        return true;
    }
    catch (e) {
        if (e.code != "ENOENT") {
            throw e;
        }
    }
    return false;
}
exports.pathExists = pathExists;
function colored(string, color, bold) {
    if (!SUPPORTS_COLOR) {
        return string;
    }
    var prefix = [];
    if (bold) {
        prefix.push(ansi(1));
    }
    if (color) {
        prefix.push(ansi(colors.indexOf(color) + 31));
    }
    return prefix.join("") + string + ansi(0);
}
exports.colored = colored;
function supports_color() {
    var stdout = process.stdout;
    if (!stdout.isTTY) {
        return false;
    }
    if (process.platform === "win32") {
        return false;
    }
    if ("COLORTERM" in process.env) {
        return true;
    }
    if (process.env.TERM === "dumb") {
        return false;
    }
    if (process.env.TERM &&
        /^screen|^xterm|^vt100|color|ansi|cygwin|linux/i.test(process.env.TERM)) {
        return true;
    }
    return false;
}
var SUPPORTS_COLOR = supports_color();
function repeat(str, num) {
    return new Array(num + 1).join(str);
}
exports.repeat = repeat;
function generators_available() {
    var gen;
    try {
        eval("gen = function *(){}"); // jshint ignore:line
        return (typeof gen === "function" && gen.constructor.name == "GeneratorFunction");
    }
    catch (e) {
        return false;
    }
}
exports.generators_available = generators_available;
function wrap(lines, width) {
    var e_1, _a;
    var ans = [];
    var prev = "";
    try {
        for (var lines_1 = __values(lines), lines_1_1 = lines_1.next(); !lines_1_1.done; lines_1_1 = lines_1.next()) {
            var line = lines_1_1.value;
            line = prev + line;
            prev = "";
            if (line.length > width) {
                prev = line.slice(width);
                if (prev) {
                    prev += " ";
                }
                line = line.slice(0, width - 1);
                if (line.slice(-1) !== " ") {
                    line += "-";
                }
            }
            ans.push(line);
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (lines_1_1 && !lines_1_1.done && (_a = lines_1.return)) _a.call(lines_1);
        }
        finally { if (e_1) throw e_1.error; }
    }
    if (prev) {
        return ans.concat(wrap([prev], width));
    }
    return ans;
}
exports.wrap = wrap;
function merge() {
    // Simple merge of properties from all objects
    var ans = {};
    Array.prototype.slice.call(arguments).forEach(function (arg) {
        Object.keys(arg).forEach(function (key) {
            ans[key] = arg[key];
        });
    });
    return ans;
}
exports.merge = merge;
function getImportDirs(paths_string, ignore_env) {
    var _a;
    var paths = [];
    function merge(new_path) {
        if (!paths.includes(new_path)) {
            paths.push(new_path);
        }
    }
    if (!ignore_env && ((_a = process === null || process === void 0 ? void 0 : process.env) === null || _a === void 0 ? void 0 : _a.JPYTHONPATH)) {
        process.env.JPYTHONPATH.split(path_1.delimiter).forEach(merge);
    }
    if (paths_string) {
        paths_string.split(path_1.delimiter).forEach(merge);
    }
    return paths;
}
exports.getImportDirs = getImportDirs;
function sha1sum(data) {
    var h = (0, crypto_1.createHash)("sha1");
    h.update(data);
    return h.digest("hex");
}
exports.sha1sum = sha1sum;
function once(obj, event) {
    return __awaiter(this, void 0, void 0, function () {
        function wait(cb) {
            obj.once(event, function () {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                cb(undefined, args);
            });
        }
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    if (!(obj instanceof events_1.EventEmitter)) {
                        // just in case typescript doesn't catch something:
                        throw Error("obj must be an EventEmitter");
                    }
                    return [4 /*yield*/, (0, util_1.promisify)(wait)()];
                case 1: return [2 /*return*/, _a.sent()];
            }
        });
    });
}
exports.once = once;
//# sourceMappingURL=utils.js.map